# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2017-2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
from unittest import mock

import fixtures
from testtools.matchers import Contains, Equals, Not

import tests
from snapcraft import storeapi
from snapcraft.storeapi.errors import StoreUploadError

from . import CommandBaseTestCase


class UploadMetadataCommandTestCase(CommandBaseTestCase):
    def setUp(self):
        super().setUp()

        self.fake_precheck = fixtures.MockPatch(
            "snapcraft.storeapi.StoreClient.upload_precheck"
        )
        self.useFixture(self.fake_precheck)

        self.fake_metadata = fixtures.MockPatchObject(
            storeapi.StoreClient, "upload_metadata"
        )
        self.useFixture(self.fake_metadata)

        self.uploaded_icon = None

        def _save_updated_icon(snap_name, metadata, force):
            self.uploaded_icon = metadata["icon"].read() if metadata["icon"] else None

        self.fake_binary_metadata = fixtures.MockPatchObject(
            storeapi.StoreClient,
            "upload_binary_metadata",
            side_effect=_save_updated_icon,
        )
        self.useFixture(self.fake_binary_metadata)

        self.snap_file = os.path.join(
            os.path.dirname(tests.__file__), "data", "test-snap-with-icon.snap"
        )

    def assert_expected_metadata_calls(self, force=False, optional_text_metadata=None):
        # text metadata
        text_metadata = {
            "description": "Description of the most simple snap",
            "summary": "Summary of the most simple snap",
        }
        if optional_text_metadata is not None:
            text_metadata.update(optional_text_metadata)

        self.fake_metadata.mock.assert_called_once_with(
            snap_name="basic", metadata=text_metadata, force=force
        )
        # binary metadata
        args, _ = self.fake_binary_metadata.mock.call_args
        self.assertEqual(args[0], "basic")
        expected_icon = (
            b'<svg width="256" height="256">\n'
            b'<rect width="256" height="256" style="fill:rgb(0,0,255)" />\n'
            b"</svg>"
        )
        self.assertEqual(self.uploaded_icon, expected_icon)
        self.assertEqual(args[2], force)

    def test_without_snap_must_raise_exception(self):
        result = self.run_command(["push-metadata"])

        self.assertThat(result.exit_code, Equals(2))
        self.assertThat(result.output, Contains("Usage:"))

    def test_simple(self):
        # upload metadata
        with mock.patch("snapcraft.storeapi._status_tracker.StatusTracker"):
            result = self.run_command(["upload-metadata", self.snap_file])
        self.assertThat(result.exit_code, Equals(0))

        self.assertThat(
            result.output,
            Not(Contains("Uploading metadata to the Store (force=False)")),
        )
        self.assertThat(result.output, Contains("The metadata has been uploaded"))
        self.assert_expected_metadata_calls(force=False)

    def test_with_license_and_title(self):
        self.snap_file = os.path.join(
            os.path.dirname(tests.__file__),
            "data",
            "test-snap-with-icon-license-title.snap",
        )

        # upload metadata
        with mock.patch("snapcraft.storeapi._status_tracker.StatusTracker"):
            result = self.run_command(["upload-metadata", self.snap_file])
        self.assertThat(result.exit_code, Equals(0))

        self.assertThat(
            result.output,
            Not(Contains("Uploading metadata to the Store (force=False)")),
        )
        self.assertThat(result.output, Contains("The metadata has been uploaded"))
        self.assert_expected_metadata_calls(
            force=False, optional_text_metadata={"title": "Basic", "license": "GPL-3.0"}
        )

    def test_simple_debug(self):
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_ENABLE_DEVELOPER_DEBUG", "yes")
        )
        # upload metadata
        with mock.patch("snapcraft.storeapi._status_tracker.StatusTracker"):
            result = self.run_command(["upload-metadata", self.snap_file])
        self.assertThat(result.exit_code, Equals(0))

        self.assertThat(
            result.output, Contains("Uploading metadata to the Store (force=False)")
        )
        self.assertThat(result.output, Contains("The metadata has been uploaded"))
        self.assert_expected_metadata_calls(force=False)

    def test_upload_metadata_without_login_must_ask(self):
        self.fake_store_login = fixtures.MockPatchObject(storeapi.StoreClient, "login")
        self.useFixture(self.fake_store_login)

        self.fake_store_account_info = fixtures.MockPatchObject(
            storeapi._dashboard_api.DashboardAPI,
            "get_account_information",
            return_value={
                "account_id": "abcd",
                "account_keys": list(),
                "snaps": {
                    "16": {
                        "snap-test": {
                            "snap-id": "snap-test-snap-id",
                            "status": "Approved",
                            "private": False,
                            "since": "2016-12-12T01:01Z",
                            "price": "0",
                        }
                    }
                },
            },
        )
        self.useFixture(self.fake_store_account_info)

        self.fake_metadata.mock.side_effect = [
            storeapi.http_clients.errors.InvalidCredentialsError("error"),
            None,
        ]

        result = self.run_command(
            ["upload-metadata", self.snap_file], input="user@example.com\nsecret\n"
        )
        self.assertThat(
            result.output, Contains("You are required to login before continuing.")
        )

    def test_nonexisting_snap_must_raise_exception(self):
        result = self.run_command(["upload-metadata", "test-unexisting-snap"])
        self.assertThat(result.exit_code, Equals(2))

    def test_unregistered_snap_must_raise_exception(self):
        class MockResponse:
            status_code = 404

            def json(self):
                return dict(
                    error_list=[
                        {
                            "code": "resource-not-found",
                            "message": "Snap not found for name=basic",
                        }
                    ]
                )

        self.fake_precheck.mock.side_effect = StoreUploadError("basic", MockResponse())

        raised = self.assertRaises(
            storeapi.errors.StoreUploadError,
            self.run_command,
            ["upload-metadata", self.snap_file],
        )

        self.assertThat(
            str(raised),
            Contains("This snap is not registered. Register the snap and try again."),
        )

    def test_forced(self):
        self.useFixture(
            fixtures.EnvironmentVariable("SNAPCRAFT_ENABLE_DEVELOPER_DEBUG", "yes")
        )

        result = self.run_command(["upload-metadata", self.snap_file, "--force"])

        self.assertThat(
            result.output, Contains("Uploading metadata to the Store (force=True)")
        )
        self.assertThat(result.output, Contains("The metadata has been uploaded"))
        self.assert_expected_metadata_calls(force=True)

    def test_snap_without_icon(self):
        snap_file = os.path.join(
            os.path.dirname(tests.__file__), "data", "test-snap.snap"
        )

        # upload metadata
        with mock.patch("snapcraft.storeapi._status_tracker.StatusTracker"):
            result = self.run_command(["upload-metadata", snap_file])
        self.assertThat(result.exit_code, Equals(0))

        self.assertThat(result.output, Contains("The metadata has been uploaded"))
        # icon uploaded to store is None
        self.assertIsNone(self.uploaded_icon)

    def test_push_raises_deprecation_warning(self):
        fake_logger = fixtures.FakeLogger(level=logging.INFO)
        self.useFixture(fake_logger)

        # upload metadata
        with mock.patch("snapcraft.storeapi._status_tracker.StatusTracker"):
            result = self.run_command(["push-metadata", self.snap_file])
        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            fake_logger.output,
            Contains(
                "DEPRECATED: The 'push' set of commands have been replaced with 'upload'."
            ),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The qmake plugin is useful for building qmake-based parts.

These are projects that are built using .pro files.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

Additionally, this plugin uses the following plugin-specific keywords:

    - qmake-parameters:
      (list of strings)
      additional options to pass to the qmake invocation.

    - qmake-project-file:
      (string)
      the qmake project file to use. This is usually only needed if
      qmake can not determine what project file to use on its own.
"""

from typing import Any, Dict, List, Set

from snapcraft.plugins.v2 import PluginV2


class QMakePlugin(PluginV2):
    @classmethod
    def get_schema(cls) -> Dict[str, Any]:
        return {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "qmake-parameters": {
                    "type": "array",
                    "uniqueItems": True,
                    "items": {"type": "string"},
                    "default": [],
                },
                "qmake-project-file": {"type": "string", "default": ""},
            },
        }

    def get_build_snaps(self) -> Set[str]:
        return set()

    def get_build_packages(self) -> Set[str]:
        return {"g++", "make", "qt5-qmake"}

    def get_build_environment(self) -> Dict[str, str]:
        return {"QT_SELECT": "qt5"}

    @property
    def out_of_source_build(self):
        return True

    def _get_qmake_configure_command(self) -> str:
        cmd = [
            "qmake",
            'QMAKE_CFLAGS+="${CFLAGS:-}"',
            'QMAKE_CXXFLAGS+="${CXXFLAGS:-}"',
            'QMAKE_LFLAGS+="${LDFLAGS:-}"',
        ] + self.options.qmake_parameters

        if self.options.qmake_project_file:
            cmd.append(
                '"${{SNAPCRAFT_PART_SRC_WORK}}/{}"'.format(
                    self.options.qmake_project_file
                )
            )
        else:
            cmd.append('"${SNAPCRAFT_PART_SRC_WORK}"')

        return " ".join(cmd)

    def get_build_commands(self) -> List[str]:
        return [
            self._get_qmake_configure_command(),
            # Avoid overriding the CFLAGS and CXXFLAGS environment
            # variables qmake sets in the generated Makefile
            'env -u CFLAGS -u CXXFLAGS make -j"${SNAPCRAFT_PARALLEL_BUILD_COUNT}"',
            'make install INSTALL_ROOT="${SNAPCRAFT_PART_INSTALL}"',
        ]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import os
import shutil

import fixtures
from testtools.matchers import Equals, FileContains, FileExists, Is

from snapcraft.internal.meta import command, errors
from tests import unit


def _create_file(file_path: str, *, mode=0o755, contents="") -> None:
    os.makedirs(os.path.dirname(file_path), exist_ok=True)
    with open(file_path, "w") as f:
        if contents:
            f.write(contents)
    os.chmod(file_path, mode)


class CommandWithoutWrapperAllowedTest(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.fake_logger = fixtures.FakeLogger(level=logging.WARNING)
        self.useFixture(self.fake_logger)

    def tearDown(self):
        super().tearDown()

        self.assertThat(self.fake_logger.output, Equals(""))

    def test_command(self):
        _create_file(os.path.join(self.path, "foo"))
        cmd = command.Command(app_name="foo", command_name="command", command="foo")

        cmd.prime_command(
            can_use_wrapper=False, massage_command=True, prime_dir=self.path,
        )

        self.assertThat(cmd.command, Equals("foo"))

    def test_command_with_args(self):
        _create_file(os.path.join(self.path, "foo"))
        cmd = command.Command(
            app_name="foo", command_name="command", command="foo bar -baz"
        )
        cmd.prime_command(
            can_use_wrapper=False, massage_command=True, prime_dir=self.path,
        )

        self.assertThat(cmd.command, Equals("foo bar -baz"))


def test_interpretered_command_from_host(monkeypatch, tmp_path):
    monkeypatch.setattr(shutil, "which", lambda x: "/bin/python3")

    _create_file(os.path.join(tmp_path, "foo"), contents="#!/usr/bin/env python3\n")
    cmd = command.Command(
        app_name="foo", command_name="command", command="foo bar -baz"
    )
    cmd.prime_command(
        can_use_wrapper=True, massage_command=True, prime_dir=tmp_path.as_posix(),
    )

    assert cmd.command == "command-foo.wrapper"
    assert cmd.wrapped_command == "/bin/python3 $SNAP/foo bar -baz"


def test_interpretered_command_from_prime(tmp_path):
    _create_file(os.path.join(tmp_path, "bin", "python3"))
    _create_file(os.path.join(tmp_path, "foo"), contents="#!/usr/bin/env python3\n")

    cmd = command.Command(
        app_name="foo", command_name="command", command="foo bar -baz"
    )
    cmd.prime_command(
        can_use_wrapper=True, massage_command=True, prime_dir=tmp_path.as_posix(),
    )

    assert cmd.command == "bin/python3 $SNAP/foo bar -baz"


def test_interpretered_command_from_root(tmp_path):
    _create_file(os.path.join(tmp_path, "foo"), contents="#!/bin/sh\n")

    cmd = command.Command(
        app_name="foo", command_name="command", command="foo bar -baz"
    )
    cmd.prime_command(
        can_use_wrapper=True, massage_command=True, prime_dir=tmp_path.as_posix(),
    )

    assert cmd.command == "foo bar -baz"


class CommandWithoutWrapperAllowedTestErrors(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.fake_logger = fixtures.FakeLogger(level=logging.WARNING)
        self.useFixture(self.fake_logger)

    def test_command_starts_with_slash(self):
        cmd = command.Command(app_name="foo", command_name="command", command="/foo")

        self.assertRaises(
            errors.InvalidAppCommandFormatError,
            cmd.prime_command,
            can_use_wrapper=False,
            massage_command=True,
            prime_dir=self.path,
        )
        self.assertThat(self.fake_logger.output, Equals(""))

    def test_command_relative_command_found_in_slash(self):
        cmd = command.Command(app_name="foo", command_name="command", command="sh")

        self.assertRaises(
            errors.InvalidAppCommandFormatError,
            cmd.prime_command,
            can_use_wrapper=False,
            massage_command=True,
            prime_dir=self.path,
        )
        self.assertThat(
            self.fake_logger.output.strip(),
            Equals(
                "The command 'sh' for 'sh' was resolved to '/bin/sh'.\n"
                "The command 'sh' has been changed to '/bin/sh'."
            ),
        )

    def test_command_does_not_match_snapd_pattern(self):
        _create_file(os.path.join(self.path, "foo"))
        cmd = command.Command(
            app_name="foo", command_name="command", command="foo /!option"
        )

        self.assertRaises(
            errors.InvalidAppCommandFormatError,
            cmd.prime_command,
            can_use_wrapper=False,
            massage_command=True,
            prime_dir=self.path,
        )
        self.assertThat(self.fake_logger.output, Equals(""))

    def test_command_not_executable(self):
        _create_file(os.path.join(self.path, "foo"), mode=0o644)
        cmd = command.Command(app_name="foo", command_name="command", command="foo")

        self.assertRaises(
            errors.InvalidAppCommandNotExecutable,
            cmd.prime_command,
            can_use_wrapper=False,
            massage_command=True,
            prime_dir=self.path,
        )
        self.assertThat(self.fake_logger.output, Equals(""))

    def test_command_not_found(self):
        cmd = command.Command(app_name="foo", command_name="command", command="foo")

        self.assertRaises(
            errors.InvalidAppCommandNotFound,
            cmd.prime_command,
            can_use_wrapper=False,
            massage_command=False,
            prime_dir=self.path,
        )
        self.assertThat(self.fake_logger.output.strip(), Equals(""))


class CommandWithWrapperTest(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.fake_logger = fixtures.FakeLogger(level=logging.WARNING)
        self.useFixture(self.fake_logger)

    def test_command(self):
        _create_file(os.path.join(self.path, "foo"))
        cmd = command.Command(app_name="foo", command_name="command", command="foo")
        cmd.prime_command(
            can_use_wrapper=True, massage_command=True, prime_dir=self.path,
        )

        self.assertThat(cmd.command, Equals("foo"))
        self.assertThat(self.fake_logger.output, Equals(""))

    def test_command_with_dollar_snap_and_does_not_match_snapd_pattern(self):
        _create_file(os.path.join(self.path, "foo"))
        cmd = command.Command(
            app_name="foo", command_name="command", command="$SNAP/foo !option"
        )
        cmd.prime_command(
            can_use_wrapper=True, massage_command=True, prime_dir=self.path,
        )

        self.assertThat(cmd.command, Equals("command-foo.wrapper"))
        self.assertThat(
            self.fake_logger.output.strip(),
            Equals(
                "Found unneeded '$SNAP/' in command '$SNAP/foo !option'.\n"
                "The command '$SNAP/foo !option' has been changed to 'foo !option'.\n"
                "A shell wrapper will be generated for command 'foo !option' "
                "as it does not conform with the command pattern expected "
                "by the runtime. Commands must be relative to the prime "
                "directory and can only consist of alphanumeric characters, "
                "spaces, and the following special characters: / . _ # : $ -"
            ),
        )

    def test_command_with_args(self):
        _create_file(os.path.join(self.path, "foo"))
        cmd = command.Command(
            app_name="foo", command_name="command", command="foo bar -baz"
        )

        cmd.prime_command(
            can_use_wrapper=True, massage_command=True, prime_dir=self.path,
        )
        wrapper_path = cmd.write_wrapper(prime_dir=self.path)

        self.expectThat(cmd.command, Equals("foo bar -baz"))
        self.expectThat(wrapper_path, Is(None))
        self.assertThat(self.fake_logger.output, Equals(""))

    def test_command_does_not_match_snapd_pattern(self):
        _create_file(os.path.join(self.path, "foo"))
        cmd = command.Command(
            app_name="foo", command_name="command", command="foo /!option"
        )

        cmd.prime_command(
            can_use_wrapper=True, massage_command=True, prime_dir=self.path,
        )
        wrapper_path = cmd.write_wrapper(prime_dir=self.path)

        self.expectThat(cmd.command, Equals("command-foo.wrapper"))
        self.assertThat(wrapper_path, FileExists())
        self.assertThat(
            wrapper_path, FileContains('#!/bin/sh\nexec $SNAP/foo /!option "$@"\n')
        )
        self.assertThat(
            self.fake_logger.output.strip(),
            Equals(
                "A shell wrapper will be generated for command 'foo /!option' "
                "as it does not conform with the command pattern expected "
                "by the runtime. Commands must be relative to the prime "
                "directory and can only consist of alphanumeric characters, "
                "spaces, and the following special characters: / . _ # : $ -"
            ),
        )

    def test_command_starts_with_slash(self):
        cmd = command.Command(app_name="foo", command_name="command", command="/foo")

        cmd.prime_command(
            can_use_wrapper=True, massage_command=True, prime_dir=self.path,
        )
        wrapper_path = cmd.write_wrapper(prime_dir=self.path)

        self.expectThat(cmd.command, Equals("command-foo.wrapper"))
        self.assertThat(wrapper_path, FileExists())
        self.assertThat(wrapper_path, FileContains('#!/bin/sh\nexec /foo "$@"\n'))
        self.assertThat(
            self.fake_logger.output.strip(),
            Equals(
                "A shell wrapper will be generated for command '/foo' "
                "as it does not conform with the command pattern expected "
                "by the runtime. Commands must be relative to the prime "
                "directory and can only consist of alphanumeric characters, "
                "spaces, and the following special characters: / . _ # : $ -"
            ),
        )

    def test_command_relative_command_found_in_slash(self):
        cmd = command.Command(app_name="foo", command_name="command", command="sh")

        cmd.prime_command(
            can_use_wrapper=True, massage_command=True, prime_dir=self.path,
        )
        wrapper_path = cmd.write_wrapper(prime_dir=self.path)

        self.expectThat(cmd.command, Equals("command-foo.wrapper"))
        self.assertThat(wrapper_path, FileExists())
        self.assertThat(wrapper_path, FileContains('#!/bin/sh\nexec /bin/sh "$@"\n'))
        self.assertThat(
            self.fake_logger.output.strip(),
            Equals(
                "The command 'sh' for 'sh' was resolved to '/bin/sh'.\n"
                "The command 'sh' has been changed to '/bin/sh'.\n"
                "A shell wrapper will be generated for command '/bin/sh' "
                "as it does not conform with the command pattern expected "
                "by the runtime. Commands must be relative to the prime "
                "directory and can only consist of alphanumeric characters, "
                "spaces, and the following special characters: / . _ # : $ -"
            ),
        )

    def test_command_not_executable(self):
        _create_file(os.path.join(self.path, "foo"), mode=0o644)
        cmd = command.Command(app_name="foo", command_name="command", command="foo")

        self.assertRaises(
            errors.InvalidAppCommandNotExecutable,
            cmd.prime_command,
            can_use_wrapper=True,
            massage_command=True,
            prime_dir=self.path,
        )
        self.assertThat(self.fake_logger.output.strip(), Equals(""))

    def test_command_not_found(self):
        cmd = command.Command(app_name="foo", command_name="command", command="foo")

        self.assertRaises(
            errors.InvalidAppCommandNotFound,
            cmd.prime_command,
            can_use_wrapper=True,
            massage_command=True,
            prime_dir=self.path,
        )
        self.assertThat(self.fake_logger.output.strip(), Equals(""))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""The autotools plugin is used for autotools based parts.

Autotools based projects are the ones that have the usual
`./configure && make && make install` instruction set.

This plugin will check for the existence of a 'configure' file, if one
cannot be found, it will run 'autoconf --install'.

This plugin uses the common plugin keywords as well as those for "sources".
For more information check the 'plugins' topic for the former and the
'sources' topic for the latter.

In addition, this plugin uses the following plugin-specific keywords:

    - autotools-configure-parameters
      (list of strings)
      configure flags to pass to the build such as those shown by running
      './configure --help'
"""

from typing import Any, Dict, List, Set

from snapcraft.plugins.v2 import PluginV2


class AutotoolsPlugin(PluginV2):
    @classmethod
    def get_schema(cls) -> Dict[str, Any]:
        return {
            "$schema": "http://json-schema.org/draft-04/schema#",
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "autotools-configure-parameters": {
                    "type": "array",
                    "uniqueItems": True,
                    "items": {"type": "string"},
                    "default": [],
                }
            },
        }

    def get_build_snaps(self) -> Set[str]:
        return set()

    def get_build_packages(self) -> Set[str]:
        return {"autoconf", "automake", "autopoint", "gcc", "libtool"}

    def get_build_environment(self) -> Dict[str, str]:
        return dict()

    def _get_configure_command(self) -> str:
        cmd = ["./configure"] + self.options.autotools_configure_parameters

        return " ".join(cmd)

    def get_build_commands(self) -> List[str]:
        return [
            "[ ! -f ./configure ] && [ -f ./autogen.sh ] && env NOCONFIGURE=1 ./autogen.sh",
            "[ ! -f ./configure ] && [ -f ./bootstrap ] && env NOCONFIGURE=1 ./bootstrap",
            "[ ! -f ./configure ] && autoreconf --install",
            self._get_configure_command(),
            'make -j"${SNAPCRAFT_PARALLEL_BUILD_COUNT}"',
            'make install DESTDIR="${SNAPCRAFT_PART_INSTALL}"',
        ]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import textwrap
from typing import Tuple

from testtools.matchers import Contains, Equals, Not

import snapcraft.yaml_utils.errors
from snapcraft.internal.project_loader import errors
from snapcraft.internal.project_loader._extensions._extension import Extension
from tests import fixture_setup

from .. import ProjectLoaderBaseTest


class ExtensionTestBase(ProjectLoaderBaseTest):
    def set_attributes(self, kwargs):
        self.__dict__.update(kwargs)

    def setUp(self):
        super().setUp()

        # Create a few fake extensions
        self.useFixture(_build_environment_extension_fixture())
        self.useFixture(_build_environment2_extension_fixture())
        self.useFixture(_environment_extension_fixture())
        self.useFixture(_plug_extension_fixture())
        self.useFixture(_plug2_extension_fixture())
        self.useFixture(_daemon_extension_fixture())
        self.useFixture(_adopt_info_extension_fixture())
        self.useFixture(_invalid_extension_fixture())


class BasicExtensionTest(ExtensionTestBase):
    def test_one_app_one_extension(self):
        config = self.make_snapcraft_project(
            textwrap.dedent(
                """\
                    name: test
                    version: "1"
                    summary: test
                    description: test
                    base: core18
                    grade: stable
                    confinement: strict

                    apps:
                        test-app:
                            command: echo "hello"
                            extensions: [environment]

                    parts:
                        part1:
                            plugin: nil
                    """
            )
        )

        # Verify that the extension was removed
        self.expectThat(config.data["apps"]["test-app"], Not(Contains("extensions")))

        # Verify that the extension took effect on the app
        self.assertThat(config.data["apps"]["test-app"], Contains("environment"))
        self.assertThat(
            config.data["apps"]["test-app"]["environment"], Contains("TEST_EXTENSION")
        )
        self.expectThat(
            config.data["apps"]["test-app"]["environment"]["TEST_EXTENSION"], Equals(1)
        )

        # Verify that the extension took effect on the part
        self.assertThat(config.parts.after_requests, Contains("part1"))
        self.expectThat(
            config.parts.after_requests["part1"], Equals(["extension-part"])
        )

        # Verify that the extension added a part
        self.assertThat(config.data["parts"], Contains("extension-part"))
        self.expectThat(config.data["parts"]["extension-part"]["plugin"], Equals("nil"))

        # Verify that the extension added a root property
        self.assertThat(config.data, Contains("environment"))
        self.expectThat(config.data["environment"], Contains("TEST_EXTENSION"))

    def test_two_apps_one_extension(self):
        config = self.make_snapcraft_project(
            textwrap.dedent(
                """\
                    name: test
                    version: "1"
                    summary: test
                    description: test
                    base: core18
                    grade: stable
                    confinement: strict

                    apps:
                        app1:
                            command: app1
                            extensions: [environment]
                        app2:
                            command: app2
                            extensions: [environment]

                    parts:
                        part1:
                            plugin: nil
                    """
            )
        )

        # Verify that the extension was removed from both apps, and took effect on both
        for app_name in ("app1", "app2"):
            self.expectThat(config.data["apps"][app_name], Not(Contains("extensions")))
            self.assertThat(config.data["apps"][app_name], Contains("environment"))
            self.assertThat(
                config.data["apps"][app_name]["environment"], Contains("TEST_EXTENSION")
            )
            self.expectThat(
                config.data["apps"][app_name]["environment"]["TEST_EXTENSION"],
                Equals(1),
            )

        # Verify that the extension took effect on the part only once
        self.assertThat(config.parts.after_requests, Contains("part1"))
        self.expectThat(
            config.parts.after_requests["part1"], Equals(["extension-part"])
        )

        # Verify that the extension added a single part
        self.assertThat(config.data["parts"], Contains("extension-part"))
        self.expectThat(config.data["parts"]["extension-part"]["plugin"], Equals("nil"))

        # Verify that the extension added a root property
        self.assertThat(config.data, Contains("environment"))
        self.expectThat(config.data["environment"], Contains("TEST_EXTENSION"))

    def test_multiple_extensions(self):
        config = self.make_snapcraft_project(
            textwrap.dedent(
                """\
                    name: test
                    version: "1"
                    summary: test
                    description: test
                    base: core18
                    grade: stable
                    confinement: strict

                    apps:
                        test-app:
                            command: test-app
                            extensions: [environment, daemon]

                    parts:
                        part1:
                            plugin: nil
                    """
            )
        )

        # Verify that both extensions were applied to the app
        self.expectThat(config.data["apps"]["test-app"], Not(Contains("extensions")))
        self.assertThat(config.data["apps"]["test-app"], Contains("environment"))
        self.assertThat(
            config.data["apps"]["test-app"]["environment"], Contains("TEST_EXTENSION")
        )
        self.expectThat(
            config.data["apps"]["test-app"]["environment"]["TEST_EXTENSION"], Equals(1)
        )
        self.assertThat(config.data["apps"]["test-app"], Contains("daemon"))
        self.expectThat(config.data["apps"]["test-app"]["daemon"], Equals("simple"))

        # Verify that the environment took effect
        self.assertThat(config.parts.after_requests, Contains("part1"))
        self.expectThat(
            config.parts.after_requests["part1"], Equals(["extension-part"])
        )

        # Verify that the environment extension added a part
        self.assertThat(config.data["parts"], Contains("extension-part"))
        self.expectThat(config.data["parts"]["extension-part"]["plugin"], Equals("nil"))

        # Verify that the environment extension added a root property
        self.assertThat(config.data, Contains("environment"))
        self.expectThat(config.data["environment"], Contains("TEST_EXTENSION"))


class ExtensionOrderConsistencyTest(ExtensionTestBase):
    def assert_extensions(self, extensions):
        config = self.make_snapcraft_project(
            textwrap.dedent(
                """\
                    name: test
                    version: "1"
                    summary: test
                    description: test
                    base: core18
                    grade: stable
                    confinement: strict

                    apps:
                        test-app:
                            command: test-app
                            extensions: {extensions}

                    parts:
                        part1:
                            plugin: nil
                    """
            ).format(extensions=extensions)
        )

        # Verify that both extensions were applied to the app
        self.expectThat(config.data["apps"]["test-app"], Not(Contains("extensions")))
        self.assertThat(config.data["apps"]["test-app"], Contains("plugs"))
        self.assertThat(
            config.data["apps"]["test-app"]["plugs"],
            Equals(["test-plug2", "test-plug"]),
        )

    def test_extension_merge_plug_plug2(self):
        self.assert_extensions("[plug, plug2]")

    def test_extension_merge_plug2_plug(self):
        self.assert_extensions("[plug2, plug]")


class ExtensionMergeTest(ExtensionTestBase):
    def run_test(self):
        snapcraft_yaml = textwrap.dedent(
            """\
            name: test
            version: "1"
            summary: test
            description: test
            base: core18
            grade: stable
            confinement: strict

            apps:
                test-app:
            {app_definition}

            parts:
                part1:
            {part_definition}
            """
        )
        config = self.make_snapcraft_project(
            snapcraft_yaml.format(
                app_definition=textwrap.indent(self.app_definition, " " * 8),
                part_definition=textwrap.indent(self.part_definition, " " * 8),
            )
        )

        # Verify that the extension was removed
        self.expectThat(config.data, Not(Contains("extensions")))
        self.expectThat(config.data["apps"]["test-app"], Not(Contains("extensions")))

        # Verify that the extension took effect on the app
        self.assertThat(
            config.data["apps"]["test-app"], Equals(self.expected_app_definition)
        )
        # Verify that the extension took effect on the part
        self.assertThat(
            config.data["parts"]["part1"], Equals(self.expected_part_definition)
        )

    def test_merge_plugs(self):
        self.set_attributes(
            {
                "app_definition": textwrap.dedent(
                    """\
                    command: echo 'hello'
                    plugs: [foo]
                    extensions: [plug]
                    """
                ),
                "expected_app_definition": {
                    "command": "echo 'hello'",
                    "plugs": ["test-plug", "foo"],
                },
                "part_definition": textwrap.dedent(
                    """\
                    plugin: nil
                    """
                ),
                "expected_part_definition": {"plugin": "nil", "prime": [], "stage": []},
            }
        )

        self.run_test()

    def test_merge_environment(self):
        self.set_attributes(
            {
                "app_definition": textwrap.dedent(
                    """\
                    command: echo 'hello'
                    extensions: [environment]
                    environment:
                      FOO: BAR
                    """
                ),
                "expected_app_definition": {
                    "command": "echo 'hello'",
                    "environment": {"FOO": "BAR", "TEST_EXTENSION": 1},
                },
                "part_definition": textwrap.dedent(
                    """\
                    plugin: nil
                    """
                ),
                "expected_part_definition": {"plugin": "nil", "prime": [], "stage": []},
            }
        )

        self.run_test()

    def test_merge_build_environment(self):
        self.set_attributes(
            {
                "app_definition": textwrap.dedent(
                    """\
                    command: echo 'hello'
                    extensions: [buildenvironment]
                    """
                ),
                "expected_app_definition": {"command": "echo 'hello'"},
                "part_definition": textwrap.dedent(
                    """\
                    plugin: nil
                    build-environment:
                      - PATH: "$PATH:/part-path"
                    """
                ),
                "expected_part_definition": {
                    "plugin": "nil",
                    "build-environment": [
                        {"PATH": "$PATH:/extension-path"},
                        {"EXTKEY": "EXTVAL"},
                        {"PATH": "$PATH:/part-path"},
                    ],
                    "prime": [],
                    "stage": [],
                },
            }
        )

        self.run_test()

    def test_merge_multiple_build_environment(self):
        self.set_attributes(
            {
                "app_definition": textwrap.dedent(
                    """\
                    command: echo 'hello'
                    extensions: [buildenvironment, buildenvironment2]
                    """
                ),
                "expected_app_definition": {"command": "echo 'hello'"},
                "part_definition": textwrap.dedent(
                    """\
                    plugin: nil
                    build-environment:
                      - PATH: "$PATH:/part-path"
                    """
                ),
                "expected_part_definition": {
                    "plugin": "nil",
                    "build-environment": [
                        {"PATH": "$PATH:/extension-path2"},
                        {"EXTKEY": "EXTVAL2"},
                        {"PATH": "$PATH:/extension-path"},
                        {"EXTKEY": "EXTVAL"},
                        {"PATH": "$PATH:/part-path"},
                    ],
                    "prime": [],
                    "stage": [],
                },
            }
        )

        self.run_test()

    def test_scalar_no_override(self):
        self.set_attributes(
            {
                "app_definition": textwrap.dedent(
                    """\
                    command: echo 'hello'
                    daemon: forking
                    extensions: [daemon]
                    """
                ),
                "expected_app_definition": {
                    "command": "echo 'hello'",
                    "daemon": "forking",
                },
                "part_definition": textwrap.dedent(
                    """\
                    plugin: nil
                    """
                ),
                "expected_part_definition": {"plugin": "nil", "prime": [], "stage": []},
            }
        )

        self.run_test()


class ExtensionRootMergeTest(ExtensionTestBase):
    def run_test(self):
        snapcraft_yaml = textwrap.dedent(
            """\
            name: test
            version: "1"
            summary: test
            description: test
            base: core18
            grade: stable
            confinement: strict

            {root_definition}

            apps:
                test-app:
                    command: test-command
                    extensions: {extensions}

            parts:
                part1:
                    plugin: nil
            """
        )
        config = self.make_snapcraft_project(
            snapcraft_yaml.format(
                root_definition=self.root_definition, extensions=self.extensions
            )
        )

        # Verify that the extension was removed
        self.expectThat(config.data, Not(Contains("extensions")))

        # Verify that the extension took effect on the root of the YAML
        for key, value in self.expected_root_definition.items():
            self.assertThat(config.data[key], Equals(value))

    def test_merge_environment(self):
        self.set_attributes(
            {
                "root_definition": textwrap.dedent(
                    """\
                    environment:
                      FOO: BAR
                    """
                ),
                "extensions": "[environment]",
                "expected_root_definition": {
                    "environment": {"FOO": "BAR", "TEST_EXTENSION": 1}
                },
            }
        )

        self.run_test()

    def test_scalars_no_override(self):
        self.set_attributes(
            {
                "root_definition": textwrap.dedent(
                    """\
                    adopt-info: "test-part"
                    """
                ),
                "extensions": "[adopt]",
                "expected_root_definition": {"adopt-info": "test-part"},
            }
        )

        self.run_test()


class InvalidExtensionTest(ExtensionTestBase):
    def test_invalid_app_extension_format(self):
        raised = self.assertRaises(
            snapcraft.yaml_utils.errors.YamlValidationError,
            self.make_snapcraft_project,
            textwrap.dedent(
                """\
                name: test
                version: "1"
                summary: test
                description: test
                base: core16
                grade: stable
                confinement: strict

                apps:
                    test-app:
                        command: echo "hello"
                        extensions: I-should-be-a-list

                parts:
                    part1:
                        plugin: nil
                """
            ),
        )

        self.assertThat(
            str(raised),
            Contains(
                "The 'extensions' property does not match the required schema: "
                "'I-should-be-a-list' is not of type 'array'"
            ),
        )

    def test_duplicate_extensions(self):
        raised = self.assertRaises(
            snapcraft.yaml_utils.errors.YamlValidationError,
            self.make_snapcraft_project,
            textwrap.dedent(
                """\
                name: test
                version: "1"
                summary: test
                description: test
                grade: stable
                confinement: strict
                base: core18

                apps:
                    test-app:
                        command: echo "hello"
                        extensions: [environment, environment]

                parts:
                    my-part:
                        plugin: nil
                """
            ),
        )

        self.assertThat(
            str(raised),
            Contains(
                "The 'extensions' property does not match the required schema: "
                "['environment', 'environment'] has non-unique elements"
            ),
        )

    def test_invalid_extension_is_validated(self):
        raised = self.assertRaises(
            snapcraft.yaml_utils.errors.YamlValidationError,
            self.make_snapcraft_project,
            textwrap.dedent(
                """\
                name: test
                version: "1"
                summary: test
                description: test
                base: core18
                grade: stable
                confinement: strict

                apps:
                    test-app:
                        command: echo "hello"
                        extensions: [invalid]

                parts:
                    part1:
                        plugin: nil
                """
            ),
        )

        self.assertThat(str(raised), Contains("'unsupported-key' was unexpected"))

    def test_non_existing_extension(self):
        self.assertRaises(
            errors.ExtensionNotFoundError,
            self.make_snapcraft_project,
            textwrap.dedent(
                """\
                name: test
                version: "1"
                summary: test
                description: test
                base: core16
                grade: stable
                confinement: strict

                apps:
                    test-app:
                        command: echo "hello"
                        extensions: [not-a-extension]

                parts:
                    part1:
                        plugin: nil
                """
            ),
        )

    def test_conflicting_part(self):
        raised = self.assertRaises(
            errors.ExtensionPartConflictError,
            self.make_snapcraft_project,
            textwrap.dedent(
                """\
                name: test
                version: "1"
                summary: test
                description: test
                base: core18
                grade: stable
                confinement: strict

                apps:
                    test-app:
                        command: echo "hello"
                        extensions: [environment]

                parts:
                    extension-part:
                        plugin: nil
                """
            ),
        )

        self.assertThat(raised.extension_name, Equals("environment"))
        self.assertThat(raised.part_name, Equals("extension-part"))

    def test_no_base(self):
        self.assertRaises(
            errors.ExtensionBaseRequiredError,
            self.make_snapcraft_project,
            textwrap.dedent(
                """\
                name: test
                version: "1"
                summary: test
                description: test
                grade: stable
                confinement: strict

                apps:
                    test-app:
                        command: echo "hello"
                        extensions: [environment]

                parts:
                    extension-part:
                        plugin: nil
                """
            ),
        )

    def test_unsupported_base(self):
        raised = self.assertRaises(
            errors.ExtensionUnsupportedBaseError,
            self.make_snapcraft_project,
            textwrap.dedent(
                """\
                name: test
                version: "1"
                summary: test
                description: test
                grade: stable
                confinement: strict
                base: unsupported

                apps:
                    test-app:
                        command: echo "hello"
                        extensions: [environment]

                parts:
                    extension-part:
                        plugin: nil
                """
            ),
        )

        self.assertThat(raised.extension_name, Equals("environment"))
        self.assertThat(raised.base, Equals("unsupported"))

    def test_unsupported_default_confinement(self):
        raised = self.assertRaises(
            errors.ExtensionUnsupportedConfinementError,
            self.make_snapcraft_project,
            textwrap.dedent(
                """\
                name: test
                version: "1"
                summary: test
                description: test
                grade: stable
                base: core18

                apps:
                    test-app:
                        command: echo "hello"
                        extensions: [environment]

                parts:
                    extension-part:
                        plugin: nil
                """
            ),
        )

        self.assertThat(raised.extension_name, Equals("environment"))
        self.assertThat(raised.confinement, Equals("devmode"))

    def test_unsupported_confinement(self):
        raised = self.assertRaises(
            errors.ExtensionUnsupportedConfinementError,
            self.make_snapcraft_project,
            textwrap.dedent(
                """\
                name: test
                version: "1"
                summary: test
                description: test
                grade: stable
                confinement: unsupported
                base: core18

                apps:
                    test-app:
                        command: echo "hello"
                        extensions: [environment]

                parts:
                    extension-part:
                        plugin: nil
                """
            ),
        )

        self.assertThat(raised.extension_name, Equals("environment"))
        self.assertThat(raised.confinement, Equals("unsupported"))


def _environment_extension_fixture():
    class ExtensionImpl(Extension):
        @staticmethod
        def get_supported_bases() -> Tuple[str, ...]:
            return ("core18",)

        @staticmethod
        def get_supported_confinement() -> Tuple[str, ...]:
            return ("strict",)

        def __init__(self, extension_name, yaml_data):
            super().__init__(extension_name=extension_name, yaml_data=yaml_data)
            self.root_snippet = {"environment": {"TEST_EXTENSION": 1}}
            self.app_snippet = {"environment": {"TEST_EXTENSION": 1}}
            self.part_snippet = {"after": ["extension-part"]}
            self.parts = {"extension-part": {"plugin": "nil"}}

    return fixture_setup.FakeExtension("environment", ExtensionImpl)


def _build_environment_extension_fixture():
    class ExtensionImpl(Extension):
        @staticmethod
        def get_supported_bases() -> Tuple[str, ...]:
            return ("core18",)

        @staticmethod
        def get_supported_confinement() -> Tuple[str, ...]:
            return ("strict",)

        def __init__(self, extension_name, yaml_data):
            super().__init__(extension_name=extension_name, yaml_data=yaml_data)
            self.root_snippet = {}
            self.app_snippet = {}
            self.part_snippet = {
                "after": ["extension-part"],
                "build-environment": [
                    {"PATH": "$PATH:/extension-path"},
                    {"EXTKEY": "EXTVAL"},
                ],
            }
            self.parts = {"extension-part": {"plugin": "nil"}}

    return fixture_setup.FakeExtension("buildenvironment", ExtensionImpl)


def _build_environment2_extension_fixture():
    class ExtensionImpl(Extension):
        @staticmethod
        def get_supported_bases() -> Tuple[str, ...]:
            return ("core18",)

        @staticmethod
        def get_supported_confinement() -> Tuple[str, ...]:
            return ("strict",)

        def __init__(self, extension_name, yaml_data):
            super().__init__(extension_name=extension_name, yaml_data=yaml_data)
            self.root_snippet = {}
            self.app_snippet = {}
            self.part_snippet = {
                "after": ["extension-part2"],
                "build-environment": [
                    {"PATH": "$PATH:/extension-path2"},
                    {"EXTKEY": "EXTVAL2"},
                ],
            }
            self.parts = {"extension-part2": {"plugin": "nil"}}

    return fixture_setup.FakeExtension("buildenvironment2", ExtensionImpl)


def _plug_extension_fixture():
    class ExtensionImpl(Extension):
        @staticmethod
        def get_supported_bases() -> Tuple[str, ...]:
            return ("core18",)

        @staticmethod
        def get_supported_confinement() -> Tuple[str, ...]:
            return ("strict",)

        def __init__(self, extension_name, yaml_data):
            super().__init__(extension_name=extension_name, yaml_data=yaml_data)
            self.app_snippet = {"plugs": ["test-plug"]}

    return fixture_setup.FakeExtension("plug", ExtensionImpl)


def _plug2_extension_fixture():
    class ExtensionImpl(Extension):
        @staticmethod
        def get_supported_bases() -> Tuple[str, ...]:
            return ("core18",)

        @staticmethod
        def get_supported_confinement() -> Tuple[str, ...]:
            return ("strict",)

        def __init__(self, extension_name, yaml_data):
            super().__init__(extension_name=extension_name, yaml_data=yaml_data)
            self.app_snippet = {"plugs": ["test-plug2"]}

    return fixture_setup.FakeExtension("plug2", ExtensionImpl)


def _daemon_extension_fixture():
    class ExtensionImpl(Extension):
        @staticmethod
        def get_supported_bases() -> Tuple[str, ...]:
            return ("core18",)

        @staticmethod
        def get_supported_confinement() -> Tuple[str, ...]:
            return ("strict",)

        def __init__(self, extension_name, yaml_data):
            super().__init__(extension_name=extension_name, yaml_data=yaml_data)
            self.app_snippet = {"daemon": "simple"}

    return fixture_setup.FakeExtension("daemon", ExtensionImpl)


def _adopt_info_extension_fixture():
    class ExtensionImpl(Extension):
        @staticmethod
        def get_supported_bases() -> Tuple[str, ...]:
            return ("core18",)

        @staticmethod
        def get_supported_confinement() -> Tuple[str, ...]:
            return ("strict",)

        def __init__(self, extension_name, yaml_data):
            super().__init__(extension_name=extension_name, yaml_data=yaml_data)
            self.root_snippet = {"adopt-info": "some-part-name"}

    return fixture_setup.FakeExtension("adopt", ExtensionImpl)


def _invalid_extension_fixture():
    class ExtensionImpl(Extension):
        @staticmethod
        def get_supported_bases() -> Tuple[str, ...]:
            return ("core18",)

        @staticmethod
        def get_supported_confinement() -> Tuple[str, ...]:
            return ("strict",)

        def __init__(self, extension_name, yaml_data):
            super().__init__(extension_name=extension_name, yaml_data=yaml_data)
            self.app_snippet = {"unsupported-key": "value"}

    return fixture_setup.FakeExtension("invalid", ExtensionImpl)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from unittest import mock

import fixtures
import pytest
from testtools.matchers import DirExists, Equals, HasLength, Not

from snapcraft.internal import errors
from snapcraft.plugins.v1 import conda
from tests import unit

from . import PluginsV1BaseTestCase


class CondaPluginBaseTest(PluginsV1BaseTestCase):

    deb_arch = None

    def setUp(self):
        super().setUp()

        if self.project.deb_arch != "amd64":
            self.skipTest("architecture is not supported by conda plugin")

        self.project._snap_meta.name = "conda-snap"

        self.fake_check_call = fixtures.MockPatch("subprocess.check_call")
        self.useFixture(self.fake_check_call)


class CondaPluginPropertiesTest(unit.TestCase):
    def test_schema(self):
        schema = conda.CondaPlugin.schema()

        properties = schema["properties"]
        for expected in [
            "conda-packages",
            "conda-python-version",
            "conda-miniconda-version",
        ]:
            self.assertTrue(
                expected in properties,
                "Expected {!r} to be included in properties".format(expected),
            )

        # Check conda-packages
        conda_packages = properties["conda-packages"]
        for expected in ["type", "default", "minitems", "uniqueItems", "items"]:
            self.assertTrue(
                expected in conda_packages,
                "Expected {!r} to be included in 'conda-packages'".format(expected),
            )

        conda_packages_type = conda_packages["type"]
        self.assertThat(
            conda_packages_type,
            Equals("array"),
            'Expected "conda-packages" "type" to be "array", but '
            'it was "{}"'.format(conda_packages_type),
        )

        conda_packages_default = conda_packages["default"]
        self.assertThat(
            conda_packages_default,
            Equals([]),
            'Expected "conda-packages" "default" to be '
            '"d[]", but it was "{}"'.format(conda_packages_default),
        )

        conda_packages_minitems = conda_packages["minitems"]
        self.assertThat(
            conda_packages_minitems,
            Equals(1),
            'Expected "conda-packages" "minitems" to be 1, but '
            "it was {}".format(conda_packages_minitems),
        )

        self.assertTrue(conda_packages["uniqueItems"])

        conda_packages_items = conda_packages["items"]
        self.assertTrue(
            "type" in conda_packages_items,
            'Expected "type" to be included in "conda-packages" ' '"items"',
        )

        conda_packages_items_type = conda_packages_items["type"]
        self.assertThat(
            conda_packages_items_type,
            Equals("string"),
            'Expected "conda-packages" "item" "type" to be '
            '"string", but it was "{}"'.format(conda_packages_items_type),
        )

        # Check conda-python-version
        conda_channel = properties["conda-python-version"]
        for expected in ["type", "default"]:
            self.assertTrue(
                expected in conda_channel,
                "Expected {!r} to be included in 'conda-python-version'".format(
                    expected
                ),
            )

        conda_channel_type = conda_channel["type"]
        self.assertThat(
            conda_channel_type,
            Equals("string"),
            'Expected "conda-python-version" "type" to be "string", but '
            'it was "{}"'.format(conda_channel_type),
        )

        conda_channel_default = conda_channel["default"]
        self.assertThat(
            conda_channel_default,
            Equals(""),
            'Expected "conda-python-version" "default" to be '
            '"latest/stable", but it was "{}"'.format(conda_channel_default),
        )

        # Check conda-miniconda-version
        conda_channel = properties["conda-miniconda-version"]
        for expected in ["type", "default"]:
            self.assertTrue(
                expected in conda_channel,
                "Expected {!r} to be included in 'conda-miniconda-version'".format(
                    expected
                ),
            )

        conda_channel_type = conda_channel["type"]
        self.assertThat(
            conda_channel_type,
            Equals("string"),
            'Expected "conda-miniconda-version" "type" to be "string", but '
            'it was "{}"'.format(conda_channel_type),
        )

        conda_channel_default = conda_channel["default"]
        self.assertThat(
            conda_channel_default,
            Equals("latest"),
            'Expected "conda-miniconda-version" "default" to be '
            '"latest", but it was "{}"'.format(conda_channel_default),
        )

        # Check required properties
        self.assertIn("required", schema)
        self.assertThat(schema["required"], Equals(["conda-packages"]))

    def test_get_pull_properties(self):
        expected_pull_properties = ["conda-miniconda-version"]
        resulting_pull_properties = conda.CondaPlugin.get_pull_properties()

        self.assertThat(
            resulting_pull_properties, HasLength(len(expected_pull_properties))
        )

        for property in expected_pull_properties:
            self.assertIn(property, resulting_pull_properties)

    def test_get_build_properties(self):
        expected_build_properties = ["conda-packages", "conda-python-version"]
        resulting_build_properties = conda.CondaPlugin.get_build_properties()

        self.assertThat(
            resulting_build_properties, HasLength(len(expected_build_properties))
        )

        for property in expected_build_properties:
            self.assertIn(property, resulting_build_properties)


@pytest.mark.parametrize(
    "miniconda_version,expected_url,expected_checksum",
    [
        (
            "latest",
            "https://repo.anaconda.com/miniconda/Miniconda3-latest-Linux-x86_64.sh",
            None,
        ),
        (
            "4.6.14",
            "https://repo.anaconda.com/miniconda/Miniconda3-4.6.14-Linux-x86_64.sh",
            "md5/718259965f234088d785cad1fbd7de03",
        ),
        (
            "4.5.14",
            "https://repo.anaconda.com/miniconda/Miniconda3-4.5.14-Linux-x86_64.sh",
            None,
        ),
    ],
)
def test_get_miniconda_source(
    project, miniconda_version, expected_url, expected_checksum
):
    class Options:
        conda_miniconda_version = miniconda_version

    plugin = conda.CondaPlugin("test-part", Options(), project)
    source_script = plugin._get_miniconda_script()

    assert source_script.source == expected_url
    assert source_script.source_checksum == expected_checksum


class CondaPluginTest(CondaPluginBaseTest):
    def test_pull(self):
        class Options:
            conda_miniconda_version = "latest"

        fake_source_script = fixtures.MockPatch("snapcraft.internal.sources.Script")
        self.useFixture(fake_source_script)

        plugin = conda.CondaPlugin("test-part", Options(), self.project)

        def download(filepath):
            self.assertThat(
                filepath, Equals(os.path.join(plugin.partdir, "miniconda.sh"))
            )
            os.makedirs(plugin.partdir)
            open(filepath, "w").close()

        fake_source_script.mock().download.side_effect = download

        plugin.pull()

        fake_source_script.mock().download.assert_called_once_with(filepath=mock.ANY)
        self.fake_check_call.mock.assert_not_called()

    def test_build_with_defaults_but_no_conda_home(self):
        class Options:
            conda_packages = ["pkg1", "pkg2"]
            conda_python_version = ""

        plugin = conda.CondaPlugin("test-part", Options(), self.project)

        plugin.build()

        self.fake_check_call.mock.assert_has_calls(
            [
                mock.call(
                    [
                        os.path.join(plugin.partdir, "miniconda.sh"),
                        "-bfp",
                        plugin._conda_home,
                    ]
                ),
                mock.call(
                    [
                        os.path.join(plugin._conda_home, "bin", "conda"),
                        "create",
                        "--prefix",
                        plugin.installdir,
                        "--yes",
                        "pkg1",
                        "pkg2",
                    ],
                    env=dict(CONDA_TARGET_PREFIX_OVERRIDE="/snap/conda-snap/current"),
                ),
            ]
        )

    def test_build_with_defaults(self):
        class Options:
            conda_packages = ["pkg1", "pkg2"]
            conda_python_version = ""

        plugin = conda.CondaPlugin("test-part", Options(), self.project)

        # Fake _conda_home
        conda_cmd = os.path.join(plugin._conda_home, "bin", "conda")
        os.makedirs(os.path.dirname(conda_cmd))
        open(conda_cmd, "w").close()

        plugin.build()

        self.fake_check_call.mock.assert_called_once_with(
            [
                os.path.join(plugin._conda_home, "bin", "conda"),
                "create",
                "--prefix",
                plugin.installdir,
                "--yes",
                "pkg1",
                "pkg2",
            ],
            env=dict(CONDA_TARGET_PREFIX_OVERRIDE="/snap/conda-snap/current"),
        )

    def test_build_with_specific_python(self):
        class Options:
            conda_packages = ["pkg1", "pkg2"]
            conda_python_version = "3.7"

        plugin = conda.CondaPlugin("test-part", Options(), self.project)

        # Fake _conda_home
        conda_cmd = os.path.join(plugin._conda_home, "bin", "conda")
        os.makedirs(os.path.dirname(conda_cmd))
        open(conda_cmd, "w").close()

        plugin.build()

        self.fake_check_call.mock.assert_called_once_with(
            [
                os.path.join(plugin._conda_home, "bin", "conda"),
                "create",
                "--prefix",
                plugin.installdir,
                "--yes",
                "python=3.7",
                "pkg1",
                "pkg2",
            ],
            env=dict(CONDA_TARGET_PREFIX_OVERRIDE="/snap/conda-snap/current"),
        )

    def test_clean_pull(self):
        class Options:
            pass

        plugin = conda.CondaPlugin("test-part", Options(), self.project)

        os.makedirs(plugin._conda_home)

        plugin.clean_pull()

        self.assertThat(plugin._conda_home, Not(DirExists()))

    def test_clean_build(self):
        class Options:
            pass

        plugin = conda.CondaPlugin("test-part", Options(), self.project)

        os.makedirs(plugin._conda_home)

        plugin.clean_build()

        # Make sure that _conda_home is still there after clean_build
        self.assertThat(plugin._conda_home, DirExists())


class CondaPluginUnsupportedBase(PluginsV1BaseTestCase):
    def setUp(self):
        super().setUp()

        self.project._snap_meta.base = "unsupported-base"

        class Options:
            pass

        self.options = Options()

    def test_unsupported_base_raises(self):
        self.assertRaises(
            errors.PluginBaseError,
            conda.CondaPlugin,
            "test-part",
            self.options,
            self.project,
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging

from snapcraft.internal.meta import command


class TestCommandMangle:
    scenarios = (
        (
            "no change",
            dict(
                command_path="foo",
                command_value="foo bar",
                expected_command="foo bar",
                expected_logs=[],
                shebang="",
                interpreter_path=None,
            ),
        ),
        (
            "no change subpath",
            dict(
                command_path="bin/foo",
                command_value="bin/foo bar",
                expected_command="bin/foo bar",
                expected_logs=[],
                shebang="",
                interpreter_path=None,
            ),
        ),
        (
            "starts with $SNAP",
            dict(
                command_path="foo",
                command_value="$SNAP/foo bar",
                expected_command="foo bar",
                expected_logs=[
                    "Found unneeded '$SNAP/' in command '$SNAP/foo bar'.",
                    "The command '$SNAP/foo bar' has been changed to 'foo bar'.",
                ],
                shebang="",
                interpreter_path=None,
            ),
        ),
        (
            "quoted environment variable",
            dict(
                command_path="foo",
                command_value='foo "$bar"',
                expected_command='foo "$bar"',
                expected_logs=[],
                shebang="",
                interpreter_path=None,
            ),
        ),
        (
            "starts with $SNAP with subpath",
            dict(
                command_path="bin/foo",
                command_value="$SNAP/bin/foo bar",
                expected_command="bin/foo bar",
                expected_logs=[
                    "Found unneeded '$SNAP/' in command '$SNAP/bin/foo bar'.",
                    "The command '$SNAP/bin/foo bar' has been changed to 'bin/foo bar'.",
                ],
                shebang="",
                interpreter_path=None,
            ),
        ),
        (
            "find bin",
            dict(
                command_path="bin/foo",
                command_value="foo bar",
                expected_command="bin/foo bar",
                expected_logs=[
                    "The command 'foo' for 'foo bar' was resolved to 'bin/foo'.",
                    "The command 'foo bar' has been changed to 'bin/foo bar'.",
                ],
                shebang="",
                interpreter_path=None,
            ),
        ),
        (
            "find sbin",
            dict(
                command_path="sbin/foo",
                command_value="foo bar",
                expected_command="sbin/foo bar",
                expected_logs=[
                    "The command 'foo' for 'foo bar' was resolved to 'sbin/foo'.",
                    "The command 'foo bar' has been changed to 'sbin/foo bar'.",
                ],
                shebang="",
                interpreter_path=None,
            ),
        ),
        (
            "find usr/bin",
            dict(
                command_path="usr/bin/foo",
                command_value="foo bar",
                expected_command="usr/bin/foo bar",
                expected_logs=[
                    "The command 'foo' for 'foo bar' was resolved to 'usr/bin/foo'.",
                    "The command 'foo bar' has been changed to 'usr/bin/foo bar'.",
                ],
                shebang="",
                interpreter_path=None,
            ),
        ),
        (
            "find usr/sbin",
            dict(
                command_path="usr/sbin/foo",
                command_value="foo bar",
                expected_command="usr/sbin/foo bar",
                expected_logs=[
                    "The command 'foo' for 'foo bar' was resolved to 'usr/sbin/foo'.",
                    "The command 'foo bar' has been changed to 'usr/sbin/foo bar'.",
                ],
                shebang="",
                interpreter_path=None,
            ),
        ),
        (
            "find in root",
            dict(
                command_path=None,
                command_value="sh bar",
                expected_command="/bin/sh bar",
                expected_logs=[
                    "The command 'sh' for 'sh bar' was resolved to '/bin/sh'.",
                    "The command 'sh bar' has been changed to '/bin/sh bar'.",
                ],
                shebang="",
                interpreter_path=None,
            ),
        ),
        (
            "find in non standard path",
            dict(
                command_path="bar/foo",
                command_value="foo bar",
                expected_command="bar/foo bar",
                expected_logs=[
                    "The command 'foo' for 'foo bar' was resolved to 'bar/foo'.",
                    "The command 'foo bar' has been changed to 'bar/foo bar'.",
                ],
                shebang="",
                interpreter_path=None,
            ),
        ),
        (
            "find in non standard path and preferred over root",
            dict(
                command_path="bar/sh",
                command_value="sh bar",
                expected_command="bar/sh bar",
                expected_logs=[
                    "The command 'sh' for 'sh bar' was resolved to 'bar/sh'.",
                    "The command 'sh bar' has been changed to 'bar/sh bar'.",
                ],
                shebang="",
                interpreter_path=None,
            ),
        ),
        (
            "relative shebang",
            dict(
                command_path="foo",
                command_value="foo bar",
                expected_command="bin/python $SNAP/foo bar",
                expected_logs=[
                    "The command 'foo bar' has been changed to 'bin/python $SNAP/foo bar' to safely account for the interpreter.",
                ],
                shebang="#!/usr/bin/env $SNAP/bin/python",
                interpreter_path="bin/python",
            ),
        ),
        (
            "root shebang",
            dict(
                command_path="bin/foo",
                command_value="bin/foo bar",
                expected_command="bin/foo bar",
                expected_logs=[],
                shebang="/bin/python",
                interpreter_path="/bin/python",
            ),
        ),
        (
            "pathless shebang",
            dict(
                command_path="bin/foo",
                command_value="bin/foo bar",
                expected_command="bin/python $SNAP/bin/foo bar",
                expected_logs=[
                    "The interpreter 'python' for 'bin/foo bar' was resolved to 'bin/python'.",
                    "The command 'bin/foo bar' has been changed to 'bin/python $SNAP/bin/foo bar' to safely account for the interpreter.",
                ],
                shebang="#!/usr/bin/env python",
                interpreter_path="bin/python",
            ),
        ),
        (
            "pathless shebang with $SNAP leading command",
            dict(
                command_path="bin/foo",
                command_value="$SNAP/bin/foo bar",
                expected_command="bin/python $SNAP/bin/foo bar",
                expected_logs=[
                    "The interpreter 'python' for '$SNAP/bin/foo bar' was resolved to 'bin/python'.",
                    "Found unneeded '$SNAP/' in command '$SNAP/bin/foo bar'.",
                    "The command '$SNAP/bin/foo bar' has been changed to 'bin/python $SNAP/bin/foo bar' to safely account for the interpreter.",
                ],
                shebang="#!/usr/bin/env python",
                interpreter_path="bin/python",
            ),
        ),
        (
            "with environment variable",
            dict(
                command_path="bin/foo",
                command_value="bin/foo bar $SNAP_DATA",
                expected_command="bin/foo bar $SNAP_DATA",
                expected_logs=[],
                shebang="",
                interpreter_path=None,
            ),
        ),
        (
            "single quotes preserved",
            dict(
                command_path="bin/foo",
                command_value="bin/foo bar '$SNAP_DATA'",
                expected_command="bin/foo bar '$SNAP_DATA'",
                expected_logs=[],
                shebang="",
                interpreter_path=None,
            ),
        ),
    )

    def test_mangle(
        self,
        caplog,
        tmp_work_path,
        command_path,
        command_value,
        expected_command,
        expected_logs,
        shebang,
        interpreter_path,
    ):
        caplog.set_level(logging.INFO)

        if command_path is not None:
            command_path = tmp_work_path / command_path
            command_path.parent.mkdir(parents=True, exist_ok=True)
            with command_path.open("w") as command_file:
                print(shebang, file=command_file)
            command_path.chmod(0o755)

        if interpreter_path is not None and not interpreter_path.startswith("/"):
            interpreter_path = tmp_work_path / interpreter_path
            interpreter_path.parent.mkdir(parents=True, exist_ok=True)
            interpreter_path.touch()
            interpreter_path.chmod(0o755)

        assert (
            command._SnapCommandResolver.resolve_snap_command_entry(
                command=command_value, prime_path=tmp_work_path
            )
            == expected_command
        )

        log_messages = [r.message for r in caplog.records]
        assert log_messages == expected_logs


def test_find_binary_not_found(tmp_path):
    assert (
        command._SnapCommandResolver.resolve_snap_command_entry(
            command="not-found", prime_path=tmp_path
        )
        is None
    )


def test_find_ambigious_command_multiple_targets(tmp_work_path, caplog):
    caplog.set_level(logging.INFO)
    command_paths = [tmp_work_path / "bin/xc", tmp_work_path / "usr/bin/xc"]
    for command_path in command_paths:
        command_path.parent.mkdir(parents=True, exist_ok=True)
        command_path.touch()
        command_path.chmod(0o755)

    assert (
        command._SnapCommandResolver.resolve_snap_command_entry(
            command="xc", prime_path=tmp_work_path
        )
        == "bin/xc"
    )

    log_messages = [r.message for r in caplog.records]
    assert log_messages == [
        "Multiple binaries matching for ambiguous command 'xc': ['bin/xc', 'usr/bin/xc']",
        "The command 'xc' for 'xc' was resolved to 'bin/xc'.",
        "The command 'xc' has been changed to 'bin/xc'.",
    ]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import mock

from testtools.matchers import Equals

from snapcraft.internal import steps

from . import LifecycleCommandsBaseTestCase


class TestPullBuildStagePrimeCommand(LifecycleCommandsBaseTestCase):
    def assert_build_provider_calls(self, step: steps.Step):
        self.fake_lifecycle_execute.mock.assert_not_called()
        if step is None:
            self.provider_mock.execute_step.assert_not_called()
        else:
            self.provider_mock.execute_step.assert_called_once_with(step)
        self.assert_clean_not_called()

    def run_test_using_defaults(self, step):
        result = self.run_command([step.name])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_get_provider_for.mock.assert_called_once_with("multipass")
        self.assert_build_provider_calls(step)
        self.provider_mock.shell.assert_not_called()

    def run_test_with_parts_specified_using_defaults(self, step):
        result = self.run_command([step.name, "part0", "part1", "part2"])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_get_provider_for.mock.assert_called_once_with("multipass")
        self.assert_build_provider_calls(step)
        self.provider_mock.shell.assert_not_called()

    def run_test_shell_using_defaults(self, step, previous_step):
        result = self.run_command([step.name, "--shell"])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_get_provider_for.mock.assert_called_once_with("multipass")
        self.assert_build_provider_calls(previous_step)
        self.provider_mock.shell.assert_called_once_with()

    def run_test_shell_after_using_defaults(self, step):
        result = self.run_command([step.name, "--shell-after"])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_get_provider_for.mock.assert_called_once_with("multipass")
        self.assert_build_provider_calls(step)
        self.provider_mock.shell.assert_called_once_with()

    def run_test_using_lxd(self, step):
        result = self.run_command([step.name, "--use-lxd"])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_get_provider_for.mock.assert_called_once_with("lxd")
        self.assert_build_provider_calls(step)
        self.provider_mock.shell.assert_not_called()

    def run_test_with_parts_specified_using_lxd(self, step):
        result = self.run_command([step.name, "--use-lxd", "part0", "part1", "part2"])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_get_provider_for.mock.assert_called_once_with("lxd")
        self.assert_build_provider_calls(step)
        self.provider_mock.shell.assert_not_called()

    def run_test_using_destructive_mode(self, step):
        result = self.run_command([step.name, "--destructive-mode"])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_get_provider_for.mock.assert_not_called()
        self.fake_lifecycle_execute.mock.assert_called_once_with(
            step, mock.ANY, tuple()
        )

    def run_test_with_parts_specified_using_destructive_mode(self, step):
        result = self.run_command(
            [step.name, "--destructive-mode", "part0", "part1", "part2"]
        )

        self.assertThat(result.exit_code, Equals(0))
        self.fake_get_provider_for.mock.assert_not_called()
        self.fake_lifecycle_execute.mock.assert_called_once_with(
            step, mock.ANY, tuple(["part0", "part1", "part2"])
        )

    def test_pull_defaults(self):
        self.run_test_using_defaults(step=steps.PULL)

    def test_pull_with_parts_specified_using_defaults(self):
        self.run_test_with_parts_specified_using_defaults(step=steps.PULL)

    def test_pull_shell_using_defaults(self):
        self.run_test_shell_using_defaults(step=steps.PULL, previous_step=None)

    def test_pull_shell_after_using_defaults(self):
        self.run_test_shell_after_using_defaults(step=steps.PULL)

    def test_pull_using_lxd(self):
        self.run_test_using_lxd(step=steps.PULL)

    def test_pull_with_parts_specified_using_lxd(self):
        self.run_test_with_parts_specified_using_lxd(step=steps.PULL)

    def test_pull_using_destructive_mode(self):
        self.run_test_using_destructive_mode(step=steps.PULL)

    def test_pull_with_parts_specified_using_destructive_mode(self):
        self.run_test_with_parts_specified_using_destructive_mode(step=steps.PULL)

    def test_build_defaults(self):
        self.run_test_using_defaults(step=steps.BUILD)

    def test_build_with_parts_specified_using_defaults(self):
        self.run_test_with_parts_specified_using_defaults(step=steps.BUILD)

    def test_build_shell_using_defaults(self):
        self.run_test_shell_using_defaults(step=steps.BUILD, previous_step=steps.PULL)

    def test_build_shell_after_using_defaults(self):
        self.run_test_shell_after_using_defaults(step=steps.BUILD)

    def test_build_using_lxd(self):
        self.run_test_using_lxd(step=steps.BUILD)

    def test_build_with_parts_specified_using_lxd(self):
        self.run_test_with_parts_specified_using_lxd(step=steps.BUILD)

    def test_build_using_destructive_mode(self):
        self.run_test_using_destructive_mode(step=steps.BUILD)

    def test_build_with_parts_specified_using_destructive_mode(self):
        self.run_test_with_parts_specified_using_destructive_mode(step=steps.BUILD)

    def test_stage_defaults(self):
        self.run_test_using_defaults(step=steps.STAGE)

    def test_stage_with_parts_specified_using_defaults(self):
        self.run_test_with_parts_specified_using_defaults(step=steps.STAGE)

    def test_stage_shell_using_defaults(self):
        self.run_test_shell_using_defaults(step=steps.STAGE, previous_step=steps.BUILD)

    def test_stage_shell_after_using_defaults(self):
        self.run_test_shell_after_using_defaults(step=steps.STAGE)

    def test_stage_using_lxd(self):
        self.run_test_using_lxd(step=steps.STAGE)

    def test_stage_with_parts_specified_using_lxd(self):
        self.run_test_with_parts_specified_using_lxd(step=steps.STAGE)

    def test_stage_using_destructive_mode(self):
        self.run_test_using_destructive_mode(step=steps.STAGE)

    def test_stage_with_parts_specified_using_destructive_mode(self):
        self.run_test_with_parts_specified_using_destructive_mode(step=steps.STAGE)

    def test_prime_defaults(self):
        self.run_test_using_defaults(step=steps.PRIME)

    def test_prime_with_parts_specified_using_defaults(self):
        self.run_test_with_parts_specified_using_defaults(step=steps.PRIME)

    def test_prime_shell_using_defaults(self):
        self.run_test_shell_using_defaults(step=steps.PRIME, previous_step=steps.STAGE)

    def test_prime_shell_after_using_defaults(self):
        self.run_test_shell_after_using_defaults(step=steps.PRIME)

    def test_prime_using_lxd(self):
        self.run_test_using_lxd(step=steps.PRIME)

    def test_prime_with_parts_specified_using_lxd(self):
        self.run_test_with_parts_specified_using_lxd(step=steps.PRIME)

    def test_prime_using_destructive_mode(self):
        self.run_test_using_destructive_mode(step=steps.PRIME)

    def test_prime_with_parts_specified_using_destructive_mode(self):
        self.run_test_with_parts_specified_using_destructive_mode(step=steps.PRIME)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pathlib
import re
from unittest import mock

import fixtures
import pytest
from simplejson.scanner import JSONDecodeError
from testtools.matchers import Contains, Equals, MatchesRegex, Not

from snapcraft import storeapi

from . import FakeStoreCommandsBaseTestCase


class LoginCommandTestCase(FakeStoreCommandsBaseTestCase):
    def test_login(self):
        # No 2fa
        self.fake_store_login.mock.side_effect = None

        result = self.run_command(["login"], input="user@example.com\nsecret\n")

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(result.output, Contains(storeapi.constants.TWO_FACTOR_WARNING))
        self.assertThat(result.output, Contains("Login successful."))
        self.fake_store_login.mock.assert_called_once_with(
            email="user@example.com",
            password=mock.ANY,
            acls=None,
            packages=None,
            channels=None,
            expires=None,
            save=True,
            config_fd=None,
        )

    def test_login_with_2fa(self):
        self.fake_store_login.mock.side_effect = [
            storeapi.http_clients.errors.StoreTwoFactorAuthenticationRequired(),
            None,
        ]

        # no exception raised.
        result = self.run_command(["login"], input="user@example.com\nsecret\n123456\n")

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output, Not(Contains(storeapi.constants.TWO_FACTOR_WARNING))
        )
        self.assertThat(result.output, Contains("Login successful."))

        self.assertThat(self.fake_store_login.mock.call_count, Equals(2))
        self.fake_store_login.mock.assert_has_calls(
            [
                mock.call(
                    email="user@example.com",
                    password="secret",
                    acls=None,
                    packages=None,
                    channels=None,
                    expires=None,
                    save=True,
                    config_fd=None,
                ),
                mock.call(
                    email="user@example.com",
                    password="secret",
                    otp="123456",
                    acls=None,
                    packages=None,
                    channels=None,
                    expires=None,
                    save=True,
                    config_fd=None,
                ),
            ]
        )

    def test_successful_login_with(self):
        self.useFixture(
            fixtures.MockPatchObject(
                storeapi.StoreClient,
                "acl",
                return_value={
                    "snap_ids": None,
                    "channels": None,
                    "permissions": None,
                    "expires": "2018-01-01T00:00:00",
                },
            )
        )
        self.fake_store_login.mock.side_effect = None

        pathlib.Path("exported-login").touch()

        result = self.run_command(["login", "--with", "exported-login"])

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(result.output, Contains("Login successful"))
        self.assertThat(
            result.output, MatchesRegex(r".*snaps:.*?No restriction", re.DOTALL)
        )
        self.assertThat(
            result.output, MatchesRegex(r".*channels:.*?No restriction", re.DOTALL)
        )
        self.assertThat(
            result.output, MatchesRegex(r".*permissions:.*?No restriction", re.DOTALL)
        )
        self.assertThat(
            result.output, MatchesRegex(r".*expires:.*?2018-01-01T00:00:00", re.DOTALL)
        )

        self.fake_store_login.mock.assert_called_once_with(
            email="",
            password="",
            acls=None,
            packages=None,
            channels=None,
            expires=None,
            save=True,
            config_fd=mock.ANY,
        )

    def test_login_failed_with_invalid_credentials(self):
        self.fake_store_login.mock.side_effect = storeapi.http_clients.errors.InvalidCredentialsError(
            "error"
        )

        with pytest.raises(
            storeapi.http_clients.errors.InvalidCredentialsError
        ) as exc_info:
            self.run_command(["login"], input="user@example.com\nbadsecret\n")

        assert (
            str(exc_info.value)
            == 'Invalid credentials: error. Have you run "snapcraft login"?'
        )

    def test_login_failed_with_store_authentication_error(self):
        self.fake_store_login.mock.side_effect = storeapi.http_clients.errors.StoreAuthenticationError(
            "error"
        )

        raised = self.assertRaises(
            storeapi.http_clients.errors.StoreAuthenticationError,
            self.run_command,
            ["login"],
            input="user@example.com\nbad-secret\n",
        )

        self.assertThat(raised.message, Equals("error"))

    def test_failed_login_with_store_account_info_error(self):
        response = mock.Mock()
        response.json.side_effect = JSONDecodeError("mock-fail", "doc", 1)
        response.status_code = 500
        response.reason = "Internal Server Error"
        self.fake_store_login.mock.side_effect = storeapi.errors.StoreAccountInformationError(
            response
        )

        with pytest.raises(storeapi.errors.StoreAccountInformationError) as exc_info:
            self.run_command(["login"], input="user@example.com\nsecret\n\n")

        assert (
            str(exc_info.value)
            == "Error fetching account information from store: 500 Internal Server Error"
        )

    def test_failed_login_with_dev_namespace_error(self):
        response = mock.Mock()
        response.status_code = 403
        response.reason = storeapi.constants.MISSING_NAMESPACE
        content = {
            "error_list": [
                {
                    "message": storeapi.constants.MISSING_NAMESPACE,
                    "extra": {"url": "http://fake-url.com", "api": "fake-api"},
                }
            ]
        }
        response.json.return_value = content
        self.fake_store_account_info.mock.side_effect = storeapi.errors.StoreAccountInformationError(
            response
        )

        with pytest.raises(storeapi.errors.NeedTermsSignedError) as exc_info:
            self.run_command(["login"], input="user@example.com\nsecret\n")

        assert (
            str(exc_info.value)
            == "Developer Terms of Service agreement must be signed before continuing: You need to set a username. It will appear in the developer field alongside the other details for your snap. Please visit http://fake-url.com and login again."
        )

    def test_failed_login_with_unexpected_account_error(self):
        # Test to simulate get_account_info raising unexpected errors.
        response = mock.Mock()
        response.status_code = 500
        response.reason = "Internal Server Error"
        content = {
            "error_list": [
                {
                    "message": "Just another error",
                    "extra": {"url": "http://fake-url.com", "api": "fake-api"},
                }
            ]
        }
        response.json.return_value = content
        self.fake_store_account_info.mock.side_effect = storeapi.errors.StoreAccountInformationError(
            response
        )

        with pytest.raises(storeapi.errors.StoreAccountInformationError) as exc_info:
            self.run_command(["login"], input="user@example.com\nsecret\n\n")

        assert (
            str(exc_info.value)
            == "Error fetching account information from store: Just another error"
        )

    def test_failed_login_with_dev_agreement_error_with_choice_no(self):
        response = mock.Mock()
        response.status_code = 403
        response.reason = storeapi.constants.MISSING_AGREEMENT
        content = {
            "error_list": [
                {
                    "message": storeapi.constants.MISSING_AGREEMENT,
                    "extra": {"url": "http://fake-url.com", "api": "fake-api"},
                }
            ]
        }
        response.json.return_value = content
        self.fake_store_account_info.mock.side_effect = storeapi.errors.StoreAccountInformationError(
            response
        )

        with pytest.raises(storeapi.errors.NeedTermsSignedError) as exc_info:
            self.run_command(["login"], input="user@example.com\nsecret\nn\n")

        assert (
            str(exc_info.value)
            == "Developer Terms of Service agreement must be signed before continuing: You must agree to the developer terms and conditions to upload snaps."
        )

    def test_failed_login_with_dev_agreement_error_with_choice_yes(self):
        response = mock.Mock()
        response.status_code = 403
        response.reason = storeapi.constants.MISSING_AGREEMENT
        content = {
            "error_list": [
                {
                    "message": storeapi.constants.MISSING_AGREEMENT,
                    "extra": {"url": "http://fake-url.com", "api": "fake-api"},
                }
            ]
        }
        response.json.return_value = content
        self.fake_store_account_info.mock.side_effect = storeapi.errors.StoreAccountInformationError(
            response
        )

        with pytest.raises(storeapi.errors.NeedTermsSignedError) as exc_info:
            self.run_command(["login"], input="user@example.com\nsecret\ny\n")

        assert (
            str(exc_info.value)
            == "Developer Terms of Service agreement must be signed before continuing: Unexpected error encountered during signing the developer terms and conditions. Please visit http://fake-url.com and agree to the terms and conditions before continuing."
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017,2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
import importlib
import logging
import sys
from pathlib import Path

import jsonschema

import snapcraft.yaml_utils.errors
from snapcraft import plugins
from snapcraft.internal import errors
from snapcraft.project import Project

logger = logging.getLogger(__name__)


def load_plugin(
    plugin_name: str,
    part_name: str,
    project: Project,
    properties,
    part_schema,
    definitions_schema,
) -> plugins.v1.PluginV1:
    local_plugins_dir = project._get_local_plugins_dir()
    if local_plugins_dir is not None:
        plugin_class = _get_local_plugin_class(
            plugin_name=plugin_name, local_plugins_dir=local_plugins_dir
        )
    if plugin_class is None:
        plugin_class = plugins.get_plugin_for_base(
            plugin_name, build_base=project._get_build_base()
        )

    if issubclass(plugin_class, plugins.v2.PluginV2):
        plugin_schema = plugin_class.get_schema()
        options = _make_options(
            part_name, part_schema, definitions_schema, properties, plugin_schema
        )
        plugin = plugin_class(part_name=part_name, options=options)
    else:
        plugin_schema = plugin_class.schema()
        _validate_pull_and_build_properties(
            plugin_name, plugin_class, part_schema, definitions_schema
        )
        options = _make_options(
            part_name, part_schema, definitions_schema, properties, plugin_schema
        )
        plugin = plugin_class(part_name, options, project)

        if project.is_cross_compiling:
            logger.debug(
                "Setting {!r} as the compilation target for {!r}".format(
                    project.deb_arch, plugin_name
                )
            )
            plugin.enable_cross_compilation()

    return plugin


def _load_compat_x_prefix(plugin_name: str, module_name: str, local_plugin_dir: str):
    compat_path = Path(local_plugin_dir, f"x-{plugin_name}.py")
    if not compat_path.exists():
        return None

    preferred_name = f"{module_name}.py"
    logger.warning(
        f"Legacy plugin name detected, please rename the plugin's file name {compat_path.name!r} to {preferred_name!r}."
    )

    spec = importlib.util.spec_from_file_location(plugin_name, compat_path)
    if spec.loader is None:
        return None

    # Prevent mypy type complaints by asserting type.
    assert isinstance(spec.loader, importlib.abc.Loader)

    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def _load_local(plugin_name: str, local_plugin_dir: str):
    module_name = plugin_name.replace("-", "_")

    module = _load_compat_x_prefix(plugin_name, module_name, local_plugin_dir)
    if module is None:
        sys.path = [local_plugin_dir] + sys.path
        logger.debug(
            f"Loading plugin module {module_name!r} with sys.path {sys.path!r}"
        )
        try:
            module = importlib.import_module(module_name)
        finally:
            sys.path.pop(0)

    return module


def _get_local_plugin_class(*, plugin_name: str, local_plugins_dir: str):
    with contextlib.suppress(ImportError):
        module = _load_local(plugin_name, local_plugins_dir)
        logger.info(f"Loaded local plugin for {plugin_name}")

        # v2 requires plugin implementation to be named "PluginImpl".
        if hasattr(module, "PluginImpl") and issubclass(
            module.PluginImpl, plugins.v2.PluginV2
        ):
            return module.PluginImpl

        for attr in vars(module).values():
            if not isinstance(attr, type):
                continue
            if not issubclass(attr, plugins.v1.PluginV1):
                continue
            if not hasattr(attr, "__module__"):
                continue
            logger.debug(
                f"Plugin attribute {attr!r} has __module__: {attr.__module__!r}"
            )
            if attr.__module__.startswith("snapcraft.plugins"):
                continue
            return attr
        else:
            raise errors.PluginError(f"unknown plugin: {plugin_name!r}")


def _validate_pull_and_build_properties(
    plugin_name, plugin, part_schema, definitions_schema
):
    merged_schema = _merged_part_and_plugin_schemas(
        part_schema, definitions_schema, plugin.schema()
    )
    merged_properties = merged_schema["properties"]

    # First, validate pull properties
    invalid_properties = _validate_step_properties(
        plugin.get_pull_properties(), merged_properties
    )

    if invalid_properties:
        raise errors.InvalidPullPropertiesError(plugin_name, list(invalid_properties))

    # Now, validate build properties
    invalid_properties = _validate_step_properties(
        plugin.get_build_properties(), merged_properties
    )

    if invalid_properties:
        raise errors.InvalidBuildPropertiesError(plugin_name, list(invalid_properties))


def _validate_step_properties(step_properties, schema_properties):
    invalid_properties = set()
    for step_property in step_properties:
        if step_property not in schema_properties:
            invalid_properties.add(step_property)

    return invalid_properties


def _make_options(
    part_name, part_schema, definitions_schema, properties, plugin_schema
):
    # Make copies as these dictionaries are tampered with
    part_schema = part_schema.copy()
    properties = properties.copy()

    plugin_schema = _merged_part_and_plugin_schemas(
        part_schema, definitions_schema, plugin_schema
    )

    try:
        jsonschema.validate(properties, plugin_schema)
    except jsonschema.ValidationError as e:
        error = snapcraft.yaml_utils.errors.YamlValidationError.from_validation_error(e)
        raise errors.PluginError(
            "properties failed to load for {}: {}".format(part_name, error.message)
        )

    return _populate_options(properties, plugin_schema)


def _merged_part_and_plugin_schemas(part_schema, definitions_schema, plugin_schema):
    plugin_schema = plugin_schema.copy()
    if "properties" not in plugin_schema:
        plugin_schema["properties"] = {}

    if "definitions" not in plugin_schema:
        plugin_schema["definitions"] = {}

    # The part schema takes precedence over the plugin's schema.
    plugin_schema["properties"].update(part_schema)
    plugin_schema["definitions"].update(definitions_schema)

    return plugin_schema


def _populate_options(properties, schema):
    class Options:
        pass

    options = Options()

    schema_properties = schema.get("properties", {})
    for key in schema_properties:
        attr_name = key.replace("-", "_")
        default_value = schema_properties[key].get("default")
        attr_value = properties.get(key, default_value)
        setattr(options, attr_name, attr_value)

    return options
